import{d as s,s as u,w as c,c as l,r as m,a as p,o as i,R as f,q as d,_ as v,O as y}from"./index-WCZ_4nWW.js";const r=n=>{const{value:e,isCustom:o}=f();if(d(n.query)&&o.value)try{const t=v.map("SavedSearchFilter",e.value,"LocationQuery");return{...n,query:t}}catch(t){console.error(t)}return!0},h=s({expose:[],beforeRouteEnter:r,beforeRouteUpdate:r}),_=s({...h,__name:"FlowRunsPageWithDefaultFilter",props:{component:{type:Function}},setup(n){const e=n,o=u(null);function t(a){return y(a)}return c(e.component,()=>{t(e.component)?e.component().then(a=>{o.value=a.default}):o.value=e.component},{immediate:!0}),(a,R)=>o.value!==null?(i(),l(m(o.value),{key:0})):p("",!0)}});export{_ as default};
