import{d,D as m,Y as _,e as n,c as s,h as c,i as k,o as a,k as v,u as t,bE as y,j as x,bF as N,bG as b,F as g,a as B}from"./index-WCZ_4nWW.js";import{u as C}from"./usePageTitle-Y7n4a-YH.js";const E=d({__name:"Notifications",setup(h){const r=m(),e=_(r.notifications.getNotifications),i=n(()=>e.response??[]),u=n(()=>e.executed&&i.value.length===0),l=n(()=>e.executed);return C("Notifications"),(w,o)=>{const f=k("p-layout-default");return a(),s(f,{class:"notifications"},{header:c(()=>[v(t(y))]),default:c(()=>[l.value?(a(),x(g,{key:0},[u.value?(a(),s(t(N),{key:0})):(a(),s(t(b),{key:1,notifications:i.value,onDelete:o[0]||(o[0]=p=>t(e).refresh()),onUpdate:o[1]||(o[1]=p=>t(e).refresh())},null,8,["notifications"]))],64)):B("",!0)]),_:1})}}});export{E as default};
