import{d as k,D as m,Y as i,e as t,c as a,h as n,i as _,o,k as f,u as s,bZ as x,j as P,b_ as h,b$ as y,F as b,a as v}from"./index-WCZ_4nWW.js";import{u as w}from"./usePageTitle-Y7n4a-YH.js";const F=k({__name:"WorkPools",setup(g){const r=m(),c={interval:3e4},e=i(r.workPools.getWorkPools,[{}],c),l=t(()=>e.response??[]),u=t(()=>e.executed&&l.value.length==0),p=t(()=>e.executed);return w("Work Pools"),(C,B)=>{const d=_("p-layout-default");return o(),a(d,{class:"work-pools"},{header:n(()=>[f(s(x))]),default:n(()=>[p.value?(o(),P(b,{key:0},[u.value?(o(),a(s(h),{key:0})):(o(),a(s(y),{key:1,onUpdate:s(e).refresh},null,8,["onUpdate"]))],64)):v("",!0)]),_:1})}}});export{F as default};
