import{d as m,D as p,Y as d,e as s,c as a,h as c,i as k,o,k as _,u as t,bm as f,j as i,bn as b,bo as x,F as y,a as B}from"./index-WCZ_4nWW.js";import{u as D}from"./usePageTitle-Y7n4a-YH.js";const F=m({__name:"Blocks",setup(h){const n=p(),e=d(n.blockDocuments.getBlockDocumentsCount),l=s(()=>e.executed&&e.response==0),u=s(()=>e.executed);return D("Blocks"),(C,g)=>{const r=k("p-layout-default");return o(),a(r,{class:"blocks"},{header:c(()=>[_(t(f))]),default:c(()=>[u.value?(o(),i(y,{key:0},[l.value?(o(),a(t(b),{key:0})):(o(),a(t(x),{key:1,onDelete:t(e).refresh},null,8,["onDelete"]))],64)):B("",!0)]),_:1})}}});export{F as default};
