import{d as f,D as i,Y as _,e as o,c as s,h as n,i as k,o as t,k as w,u as a,aF as y,j as C,aG as h,aH as F,F as v,a as x}from"./index-WCZ_4nWW.js";import{u as D}from"./usePageTitle-Y7n4a-YH.js";const Y=f({__name:"Flows",setup(b){const l=i(),c={interval:3e4},e=_(l.flows.getFlowsCount,[{}],c),r=o(()=>e.response??0),u=o(()=>e.executed&&r.value===0),p=o(()=>e.executed),d=()=>{e.refresh()};return D("Flows"),(g,B)=>{const m=k("p-layout-default");return t(),s(m,{class:"flows"},{header:n(()=>[w(a(y))]),default:n(()=>[p.value?(t(),C(v,{key:0},[u.value?(t(),s(a(h),{key:0})):(t(),s(a(F),{key:1,selectable:"",onDelete:d}))],64)):x("",!0)]),_:1})}}});export{Y as default};
