import{d as _,D as k,c as d,h as a,i as h,a1 as m,o as f,k as t,u,bh as w,bi as Q,ae as s,b1 as c}from"./index-WCZ_4nWW.js";import{u as C}from"./usePageTitle-Y7n4a-YH.js";const W=_({__name:"WorkQueueCreate",setup(b){const n=k(),o=m(),p=()=>{o.push(s.workQueues())},i=async r=>{try{const{id:e}=await n.workQueues.createWorkQueue(r);c("Work queue has been created","success"),o.push(s.workQueue(e))}catch(e){c("Error occurred while creating new work queue","error"),console.error(e)}};return C("Create Work Queue"),(r,e)=>{const l=h("p-layout-default");return f(),d(l,null,{header:a(()=>[t(u(w))]),default:a(()=>[t(u(Q),{action:"Create",onSubmit:i,onCancel:p})]),_:1})}}});export{W as default};
