import{d as m,D as y,b as _,Y as k,e as o,g as s,c as t,h as n,i as d,o as e,u as c,z as v,a as r,l as g,j as C,x as h,F as x}from"./index-WCZ_4nWW.js";import{u as K}from"./usePageTitle-Y7n4a-YH.js";const D=m({__name:"ArtifactKey",setup(B){const u=y(),l=_("artifactKey"),i=k(u.artifacts.getArtifactCollection,[l]),a=o(()=>i.response),f=o(()=>a.value?`${s.info.artifact}: ${a.value.key}`:s.info.artifact);return K(f),(b,w)=>{const p=d("p-layout-default");return e(),t(p,{class:"artifact"},{header:n(()=>[a.value?(e(),t(c(v),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0)]),default:n(()=>[a.value?(e(),t(c(g),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0),a.value?(e(),C(x,{key:1},[a.value.key?(e(),t(c(h),{key:0,"artifact-key":a.value.key},null,8,["artifact-key"])):r("",!0)],64)):r("",!0)]),_:1})}}});export{D as default};
