import{d as b,D as d,b as i,e as o,az as _,c as f,h as c,a as D,i as v,a1 as B,o as h,k as u,u as a,bA as w,bB as C,ae as g}from"./index-WCZ_4nWW.js";import{u as x}from"./usePageTitle-Y7n4a-YH.js";const S=b({__name:"BlockView",setup(y){const n=d(),s=B(),t=i("blockDocumentId"),l=o(()=>t.value?[t.value]:null),r=_(n.blockDocuments.getBlockDocument,l),e=o(()=>r.response),m=()=>{s.push(g.blocks())},k=o(()=>e.value?`Block: ${e.value.name}`:"Block");return x(k),(V,A)=>{const p=v("p-layout-default");return e.value?(h(),f(p,{key:0,class:"block-view"},{header:c(()=>[u(a(w),{"block-document":e.value,onDelete:m},null,8,["block-document"])]),default:c(()=>[u(a(C),{"block-document":e.value},null,8,["block-document"])]),_:1})):D("",!0)}}});export{S as default};
