import{d as k,D as b,b as i,e as o,az as y,c as m,h as t,a as _,i as d,o as f,k as l,u as s,bs as T,bt as v}from"./index-WCZ_4nWW.js";import{u as g}from"./usePageTitle-Y7n4a-YH.js";const x=k({__name:"BlocksCatalogView",setup(B){const c=b(),a=i("blockTypeSlug"),u=o(()=>a.value?[a.value]:null),n=y(c.blockTypes.getBlockTypeBySlug,u),e=o(()=>n.response),p=o(()=>e.value?`Block Type: ${e.value.name}`:null);return g(p),(C,S)=>{const r=d("p-layout-default");return e.value?(f(),m(r,{key:0,class:"blocks-catalog-view"},{header:t(()=>[l(s(T),{"block-type":e.value},null,8,["block-type"])]),default:t(()=>[l(s(v),{"block-type":e.value},null,8,["block-type"])]),_:1})):_("",!0)}}});export{x as default};
