import{d as l,b as n,c as s,h as a,i as u,o as p,k as t,u as o,c7 as c,c8 as m}from"./index-WCZ_4nWW.js";import{u as _}from"./usePageTitle-Y7n4a-YH.js";const C=l({__name:"WorkPoolQueueCreate",setup(k){const e=n("workPoolName");return _("Create Work Pool Queue"),(d,f)=>{const r=u("p-layout-default");return p(),s(r,null,{header:a(()=>[t(o(c),{"work-pool-name":o(e)},null,8,["work-pool-name"])]),default:a(()=>[t(o(m),{"work-pool-name":o(e)},null,8,["work-pool-name"])]),_:1})}}});export{C as default};
