import{d as D,D as R,b as h,Y as N,e as p,f as T,B as V,g as _,K as Y,c as n,h as e,i as o,o as l,u as a,Z as $,a as i,L as d,j as I,k as s,l as K,m as L,n as v,p as O,t as Z,S as m,v as j}from"./index-WCZ_4nWW.js";import{u as E}from"./usePageTitle-Y7n4a-YH.js";const H={key:0},F=D({__name:"Artifact",setup(P){const b=R(),k=h("artifactId"),w=N(b.artifacts.getArtifact,[k]),t=p(()=>w.response),c=T(!1),y=[{label:"Artifact"},{label:"Details"},{label:"Raw"}],u=V("tab","Artifact"),g=p(()=>t.value?`${_.info.artifact}: ${t.value.key??Y(t.value.type)}`:_.info.artifact);return E(g),(U,r)=>{const x=o("p-divider"),B=o("p-button"),C=o("p-content"),S=o("p-tabs"),A=o("p-layout-well");return l(),n(A,{class:"artifact"},{header:e(()=>[t.value?(l(),n(a($),{key:0,artifact:t.value},null,8,["artifact"])):i("",!0)]),well:e(()=>[t.value?(l(),n(a(d),{key:0,artifact:t.value,alternate:""},null,8,["artifact"])):i("",!0)]),default:e(()=>[t.value?(l(),I("section",H,[s(a(K),{artifact:t.value},null,8,["artifact"]),s(x),a(L).xl?(l(),n(C,{key:0},{default:e(()=>[s(a(v),{artifact:t.value},null,8,["artifact"]),s(B,{class:"artifact__raw-data-button",small:"",onClick:r[0]||(r[0]=f=>c.value=!c.value)},{default:e(()=>[O(Z(c.value?"Hide":"Show")+" raw data ",1)]),_:1}),c.value?(l(),n(a(m),{key:0,artifact:t.value},null,8,["artifact"])):i("",!0)]),_:1})):(l(),n(S,{key:1,selected:a(u),"onUpdate:selected":r[1]||(r[1]=f=>j(u)?u.value=f:null),tabs:y},{artifact:e(()=>[s(a(v),{artifact:t.value},null,8,["artifact"])]),details:e(()=>[s(a(d),{artifact:t.value},null,8,["artifact"])]),raw:e(()=>[s(a(m),{artifact:t.value},null,8,["artifact"])]),_:1},8,["selected"]))])):i("",!0)]),_:1})}}});export{F as default};
