import{d as i,D as f,Y as _,e as a,c as u,h as r,i as v,o,k as x,u as s,b8 as w,j as y,b9 as Q,ba as b,F as g,a as C}from"./index-WCZ_4nWW.js";import{u as h}from"./usePageTitle-Y7n4a-YH.js";const F=i({__name:"WorkQueues",setup(q){const l=f(),p={interval:3e4},e=_(l.workQueues.getWorkQueues,[{}],p),n=a(()=>e.response??[]),c=a(()=>e.executed&&n.value.length==0),k=a(()=>e.executed);return h("Work Queues"),(B,t)=>{const d=v("p-layout-default");return o(),u(d,{class:"queues"},{header:r(()=>[x(s(w))]),default:r(()=>[k.value?(o(),y(g,{key:0},[c.value?(o(),u(s(Q),{key:0})):(o(),u(s(b),{key:1,"work-queues":n.value,onUpdate:t[0]||(t[0]=m=>s(e).refresh()),onDelete:t[1]||(t[1]=m=>s(e).refresh())},null,8,["work-queues"]))],64)):C("",!0)]),_:1})}}});export{F as default};
