import{d as p,D as u,b as _,bj as k,c as m,h as l,i,o as w,k as r,u as e,c2 as d,c3 as f}from"./index-WCZ_4nWW.js";import{u as P}from"./usePageTitle-Y7n4a-YH.js";const B=p({__name:"WorkPoolEdit",async setup(C){let o,a;const s=u(),n=_("workPoolName"),t=([o,a]=k(()=>s.workPools.getWorkPoolByName(n.value)),o=await o,a(),o);return P("Create Work Pool"),(h,y)=>{const c=i("p-layout-default");return w(),m(c,null,{header:l(()=>[r(e(d),{"work-pool":e(t)},null,8,["work-pool"])]),default:l(()=>[r(e(f),{"work-pool":e(t)},null,8,["work-pool"])]),_:1})}}});export{B as default};
