import{d,f as m,D as _,e as o,Y as b,g as v,c as r,h as c,i as k,o as t,k as h,u as n,bS as y,j as C,bT as g,bU as x,F as T,a as V}from"./index-WCZ_4nWW.js";import{u as B}from"./usePageTitle-Y7n4a-YH.js";const U=d({__name:"Variables",setup(D){const l=m(),a=()=>{var e;s.value.refresh(),(e=l.value)==null||e.refreshSubscriptions()},u=_(),s=o(()=>b(u.variables.getVariables)),i=o(()=>{var e;return s.value.executed&&((e=s.value.response)==null?void 0:e.length)===0}),p=o(()=>s.value.executed);return B(v.info.variables),(e,S)=>{const f=k("p-layout-default");return t(),r(f,{class:"variables"},{header:c(()=>[h(n(y),{onCreate:a})]),default:c(()=>[p.value?(t(),C(T,{key:0},[i.value?(t(),r(n(g),{key:0,onCreate:a})):(t(),r(n(x),{key:1,ref_key:"table",ref:l,onDelete:a,onUpdate:a},null,512))],64)):V("",!0)]),_:1})}}});export{U as default};
