import{d as R,D as h,b as w,e as i,m as x,B,al as g,Y as S,c as n,h as c,i as u,a1 as T,o as s,u as t,bK as I,a as r,bL as m,k as N,bM as V,v as Y,ae as A}from"./index-WCZ_4nWW.js";import{u as K}from"./usePageTitle-Y7n4a-YH.js";const $=R({__name:"ConcurrencyLimit",setup(M){const p=h(),y=w("concurrencyLimitId"),v=T(),d=i(()=>[{label:"Details",hidden:x.xl},{label:"Active Task Runs"}]),a=B("tab","Details"),{tabs:b}=g(d,a),_={interval:3e5},L=S(p.concurrencyLimits.getConcurrencyLimit,[y.value],_),e=i(()=>L.response);function f(){v.push(A.concurrencyLimits())}const k=i(()=>e.value?`Concurrency Limit: ${e.value.tag}`:"Concurrency Limit");return K(k),(O,l)=>{const C=u("p-tabs"),D=u("p-layout-well");return s(),n(D,{class:"concurrencyLimit"},{header:c(()=>[e.value?(s(),n(t(I),{key:0,"concurrency-limit":e.value,onDelete:f},null,8,["concurrency-limit"])):r("",!0)]),well:c(()=>[e.value?(s(),n(t(m),{key:0,alternate:"","concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),default:c(()=>[N(C,{selected:t(a),"onUpdate:selected":l[0]||(l[0]=o=>Y(a)?a.value=o:null),tabs:t(b)},{details:c(()=>[e.value?(s(),n(t(m),{key:0,"concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),"active-task-runs":c(()=>{var o;return[(o=e.value)!=null&&o.activeSlots?(s(),n(t(V),{key:0,"active-slots":e.value.activeSlots},null,8,["active-slots"])):r("",!0)]}),_:1},8,["selected","tabs"])]),_:1})}}});export{$ as default};
