import{d,D as i,Y as y,e as a,c as o,h as n,i as _,o as t,k as f,u as s,aO as k,j as D,aP as h,aQ as v,F as x,a as g}from"./index-WCZ_4nWW.js";import{u as C}from"./usePageTitle-Y7n4a-YH.js";const P=d({__name:"Deployments",setup(B){const l=i(),c={interval:3e4},e=y(l.deployments.getDeployments,[{}],c),p=a(()=>e.response??[]),r=a(()=>e.executed&&p.value.length===0),u=a(()=>e.executed);return C("Deployments"),(b,F)=>{const m=_("p-layout-default");return t(),o(m,{class:"deployments"},{header:n(()=>[f(s(k))]),default:n(()=>[u.value?(t(),D(x,{key:0},[r.value?(t(),o(s(h),{key:0})):(t(),o(s(v),{key:1,onDelete:s(e).refresh},null,8,["onDelete"]))],64)):g("",!0)]),_:1})}}});export{P as default};
