import{d as w,bN as d,B as C,c as M,h as a,i as u,o as y,k as l,u as o,X as r,bO as h,v as b,bP as x,bQ as R,bR as v}from"./index-WCZ_4nWW.js";const B=w({__name:"ConcurrencyLimits",setup(G){const{showModal:t,open:p}=d(),{showModal:e,open:m}=d(),_=[{label:"Global"},{label:"Task Run"}],c=C("tab","Global");return(T,s)=>{const i=u("p-button"),k=u("p-tabs"),f=u("p-layout-default");return y(),M(f,{class:"concurrency-limits"},{header:a(()=>[l(o(r),{crumbs:[{text:"Concurrency"}]})]),default:a(()=>[l(k,{selected:o(c),"onUpdate:selected":s[2]||(s[2]=n=>b(c)?c.value=n:null),tabs:_},{global:a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Global Concurrency Limits"}]},{"after-crumbs":a(()=>[l(i,{small:"",icon:"PlusIcon",onClick:o(p)},null,8,["onClick"])]),_:1}),l(o(h),{showModal:o(t),"onUpdate:showModal":s[0]||(s[0]=n=>b(t)?t.value=n:null)},null,8,["showModal"]),l(o(x),{class:"concurrency-limits__global-table"})]),"task-run":a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Task Run Concurrency Limits"}]},{"after-crumbs":a(()=>[l(i,{small:"",icon:"PlusIcon",onClick:o(m)},null,8,["onClick"])]),_:1}),l(o(R),{showModal:o(e),"onUpdate:showModal":s[1]||(s[1]=n=>b(e)?e.value=n:null)},null,8,["showModal"]),l(o(v),{class:"concurrency-limits__task-limits-table"})]),_:1},8,["selected"])]),_:1})}}});export{B as default};
