import{d as m,D as c,b as l,bj as w,c as _,h as r,i,o as d,k as n,u as o,c9 as P,ca as f}from"./index-WCZ_4nWW.js";import{u as Q}from"./usePageTitle-Y7n4a-YH.js";const C=m({__name:"WorkPoolQueueEdit",async setup(N){let e,u;const s=c(),a=l("workPoolName"),t=l("workPoolQueueName"),p=([e,u]=w(()=>s.workPoolQueues.getWorkPoolQueueByName(a.value,t.value)),e=await e,u(),e);return Q("Edit Work Pool Queue"),(h,q)=>{const k=i("p-layout-default");return d(),_(k,null,{header:r(()=>[n(o(P),{"work-pool-name":o(a),"work-pool-queue-name":o(t)},null,8,["work-pool-name","work-pool-queue-name"])]),default:r(()=>[n(o(f),{"work-pool-name":o(a),"work-pool-queue":o(p)},null,8,["work-pool-name","work-pool-queue"])]),_:1})}}});export{C as default};
