import{d as k,D as f,f as y,e as o,bp as d,Y as m,c as _,h as s,i as v,o as C,k as t,u as l,bq as g,br as B}from"./index-WCZ_4nWW.js";import{u as T}from"./usePageTitle-Y7n4a-YH.js";const D=k({__name:"BlocksCatalog",setup(h){const c=f(),a=y(null),p=o(()=>a.value?[a.value]:[]),{filter:n}=d({blockSchemas:{blockCapabilities:p}}),i=m(c.blockTypes.getBlockTypes,[n]),u=o(()=>i.response??[]);return T("Blocks Catalog"),(x,e)=>{const r=v("p-layout-default");return C(),_(r,{class:"blocks-catalog"},{header:s(()=>[t(l(g))]),default:s(()=>[t(l(B),{capability:a.value,"onUpdate:capability":e[0]||(e[0]=b=>a.value=b),"block-types":u.value},null,8,["capability","block-types"])]),_:1})}}});export{D as default};
