import{d as p,D as l,c as _,h as e,i as m,o as d,k as o,u as n,bH as h,bI as C,ae as i,b2 as c,b1 as b}from"./index-WCZ_4nWW.js";import{u as k}from"./usePageTitle-Y7n4a-YH.js";const x=p({__name:"NotificationCreate",setup(w){const r=l();async function s(t){try{await r.notifications.createNotification(t),c.push(i.notifications())}catch(a){b("Error creating notification","error"),console.warn(a)}}function u(){c.push(i.notifications())}return k("Create Notification"),(t,a)=>{const f=m("p-layout-default");return d(),_(f,{class:"notification-create"},{header:e(()=>[o(n(h))]),default:e(()=>[o(n(C),{action:"Create",onSubmit:s,onCancel:u})]),_:1})}}});export{x as default};
